@@ -15,7 +15,7 @@
  *  Free Software Foundation, Inc.,
  *  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  *
- *  $Id: error.h 1 2006-08-26 03:31:31Z ap2c $
+ *  $Id: error.h 78 2007-10-14 12:12:08Z ap2c $
  * 
  */
 #ifndef ERROR_H_
@@ -23,12 +23,17 @@
 
 #include <stdio.h>
 
-#define _(X)			X
+#define _(X)            X
 
-#define ERROR(X, args ...)	fprintf(stderr, "[33m" X "[0m\n", ##args);
-//#define DEBUG(X, args ...)	fprintf(stderr, "[32m" X "[0m\n", ##args);
-#define DEBUG(X, args ...)
+#define ERROR(X, args ...)  fprintf(stderr, "[33m" X "[0m\n", ##args);
+#ifdef ENABLE_DEBUG
+#   define DEBUG(X, args ...)   fprintf(stderr, "[32m" X "[0m\n", ##args);
+#else
+#   define DEBUG(X, args ...)
+#endif /* ENABLE_DEBUG */
 
 
 #endif /* ERROR_H_ */
 
+/* vim: set expandtab tabstop=4 shiftwidth=4 smarttab tw=80 cin enc=utf8: */
+
